unit tablegen;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, StdCtrls, utils, utils2, ExtCtrls, math;

type
  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    OpenDialog2: TOpenDialog;
    SaveDialog2: TSaveDialog;
    Memo1: TMemo;
    MainMenu1: TMainMenu;
    Loadtxtfile1: TMenuItem;
    Loadentrypointtable1: TMenuItem;
    saveentrypointlist1: TMenuItem;
    saveentrypointtable1: TMenuItem;
    Image1: TImage;
    Image2: TImage;
    ListBox1: TListBox;
    Memo2: TMemo;
    procedure Loadtxtfile1Click(Sender: TObject);
    procedure saveentrypointlist1Click(Sender: TObject);
    procedure saveentrypointtable1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ListBox1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses tablegenUnit2;

{$R *.DFM}

const nbcrc = 128;

procedure TForm1.Loadtxtfile1Click(Sender: TObject);
var
  i: integer;
begin
  openDialog1.fileName:= '*.a';
  if opendialog1.Execute then
    with TStringList.create do
      try
        LoadFromFile(OpenDialog1.fileName);
        i:= 0;
        while i<count do
          if (length(strings[i])=0) or (strings[i][1]<>'=') then
            delete(i)
          else
            inc(i);
        memo1.text:= text;
      finally
        free;
      end;
end;

procedure TForm1.saveentrypointlist1Click(Sender: TObject);
begin
  if saveDialog1.execute then
    memo1.Lines.SaveToFile(saveDialog1.fileName);
end;

type
  Tentry = class
    value: integer;
    text: string[33];
    crc: integer;
    pos: integer;
    textPos: integer;
    constructor create(s: string; i: integer);
  end;

constructor Tentry.create(s: string; i: integer);
  function StringCRC(s: string): integer;
  var
    i: integer;
  begin
    result:= 0;
    for i:= 1 to length(s) do
      result:= result+ord(s[i]);
  end;
Begin
  text:= copy(s, 2, FindWhite(s, 1)-2);
  value:= HexToStr(copy(GetToken(s, 3), 2, maxint));
  if text[1]='^' then
    value:= ((value div 256) and $ffff0) + (value and $f);
  crc:= StringCRC(text) and (nbcrc-1);
  pos:= 0;
  textPos:= i;
end;

type
  tsortproc = function (e1, e2: TEntry): integer;

function sorts2(e1, e2: TEntry): integer;
begin
  if e1.text<e2.text then
    result := -1
  else
    if e1.text>e2.text then
      result := 1
    else
      result:= 0;
end;
function sorts3(e1, e2: TEntry): integer;
begin
  result:= e1.value-e2.value;
end;
function sortcrc(e1, e2: TEntry): integer;
begin
  result:= length(e1.text)-length(e2.text);
  if result = 0 then
    result:= sorts2(e1, e2);
end;
var
  crc: array [0..nbcrc-1] of array of TEntry; // crc: list sorted by CRC, namesize, name
procedure TForm1.saveentrypointtable1Click(Sender: TObject);

  procedure sort(var s: array of tentry; L, R: integer; sp: tsortproc);
  var
    I, J: Integer;
    e, e2: TEntry;
  begin
    repeat
      I := L;
      J := R;
      e := s[(L + R) shr 1];
      repeat
        while sp(s[I],e)<0 do Inc(I);
        while sp(s[J],e)>0 do Dec(J);
        if I <= J then
        begin
          e2:= s[I];
          s[I]:= s[J];
          s[J]:= e2;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then Sort(s, L, J, sp);
      L := I;
    until I >= R;
  end;
const
  Colors: array[1..42 {sic!}] of TColor = (clBlack, clMaroon, clGreen, clOlive,
    clNavy, clPurple, clTeal, clDkGray, clLtGray, clRed, clLime,
    clYellow, clBlue, clFuchsia, clAqua, clWhite, clScrollBar,
    clBackground, clActiveCaption, clInactiveCaption, clMenu, clWindow,
    clWindowFrame, clMenuText, clWindowText, clCaptionText,
    clActiveBorder, clInactiveBorder, clAppWorkSpace, clHighlight,
    clHighlightText, clBtnFace, clBtnShadow, clGrayText, clBtnText,
    clInactiveCaptionText, clBtnHighlight, cl3DDkShadow, cl3DLight,
    clInfoText, clInfoBk, clNone);
{
  The table generated has this structure:

  10 nibbles: strobj prolog and size
  5 nibbles: nb of entry in the table
  5 nibbles: offset to the data part of the table
  128*5 nibbles: offset to the list of entry of crc n
                (where n is the number of this entry in the 64 entry table]
  n*5 nibbles: offsets to the value of the entry n (sorted by address]
  128*
    n*5 nibbles offset to the entries of same crc. sorted by name lenght then by name
    00000
  data part
  n*
  5 nibbles: entry value
  2 nibbles: entry size in chr
  m*2 nibbles: entry text
  7 nibbles: 0000000 (end of list)
}
var
  s1, s2, s3: array of Tentry;           // s1: original list, s2: list sorted by name, s3: list sorted by adr
  i, j, k, p, m, l, nbext: integer;
  f: THPStream;                          // the output stream
  debdata: integer;                      // start of data area (after index)
  crcd: array [0..nbcrc-1] of double;
  s: TStringList;
begin
  if saveDialog2.execute then
  Begin
    s:= TStringList.Create;
    f:= THPStream.create;
    SetLength(s1, memo1.lines.count);
    SetLength(s2, memo1.lines.count);
    SetLength(s3, memo1.lines.count);
    nbext:= memo1.lines.count;

    // fill the 3+nbcrc table with the datas
    caption := 'load list'; application.processMessages;
    for i:= 0 to nbext-1 do
    Begin
      s1[i]:= TEntry.Create(memo1.lines[i], i);
      s2[i]:= s1[i];
      s3[i]:= s1[i];
      setLength(crc[s1[i].crc], length(crc[s1[i].crc])+1);
      crc[s1[i].crc][high(crc[s1[i].crc])]:= s1[i];
    end;

    // sort the 2+nbcrc tables
    caption := 'sort lists'; application.processMessages;
    Sort(s2, 0, nbext-1, sorts2);
    for i:= 0 to nbcrc-1 do
      Sort(crc[i], 0, high(crc[i]), sortcrc);
    sort(s3, 0, nbext-1, sorts3);

    // remove duplicates in the sorted by adr list
    i:= 0; j:= 0;
    while j<nbext do
      if (j<nbext-1) and (s3[j].value=s3[j+1].value) then
      begin
        s.clear;
        s.add(s3[j].text);
        while (j<nbext) and (s3[j].value=s3[j+1].value) do
          begin inc(j); s.add(s3[j].text); end;
        s3[i]:= s3[j+1-s.count+form2.Execute(s)];
        inc(i); inc(j);
      end else begin
        s3[i]:= s3[j];
        inc(i); inc(j);
      end;
    Application.MessageBox(pchar('There is '+inttostr(j-i)+' entry discarded'), 'information', mb_ok);
    for i:=i to nbext-1 do s3[i]:= s3[nbext-1];

    // write header
    caption := 'write header'; application.processMessages;
    f.WriteInt(nbext);            // nb entries
    f.WriteInt(nbcrc*10+nbext*10+5); // offset to datas
    for i:= 1 to nbcrc do            // nbcrc offset to list by CRC
      f.writeInt(0, 5);

    // write empty pointers
    caption := 'write dummy pointers'; application.processMessages;
    for i:= 1 to nbcrc+nbext*2 do
      f.writeInt(0, 5);           // empty the index tables

    // start the data area
    debData:= f.position;

    // write each entry by alphabetical order
    caption := 'write database'; application.processMessages;
    for i:= 0 to nbext-1 do
    begin
      f.WriteInt(s2[i].value);              // write the value
      s2[i].pos:= f.position;               // the value is 5 nibble before the official entry pos
      f.WriteInt(length(s2[i].text), 2);    // write the length
      for j:= 1 to length(s2[i].text) do
        f.writeInt(ord(s2[i].text[j]), 2);  // write the chrs
    end;
    f.writeInt(0, 7);

    // write the index sorted by adr
    caption := 'write adr index'; application.processMessages;
    for i:=0 to nbext-1 do
    Begin
      f.PokeInt(s3[i].pos-(nbcrc*5+10+i*5)-5, nbcrc*5+10+i*5, 5); // write each offset
      image1.Canvas.Pixels[(image1.width*i) div nbext, image1.Height-(s3[i].value*image1.Height) div (1 shl 20)]:= clBlack;
    end;

    // and by crc/size/name
    caption := 'write crc index'; application.processMessages;
    k:= nbcrc*5+nbext*5+10;
    for i:=0 to nbcrc-1 do
    begin
      f.pokeInt(k-(i*5+10), i*5+10, 5);      // write the offset to list of crc i
      for j:= 0 to high(crc[i]) do         // write each pointer
          f.PokeInt(crc[i][j].pos-(k+j*5), k+j*5, 5);
      inc(k, length(crc[i])*5+5);          // increment position pointer
    end;

    // end graphical summary
    j:= 0;
    for i:= 0 to nbcrc-1 do
      if length(crc[i])>j then
        j:= length(crc[i]);
    for i:= 0 to nbcrc-1 do
    begin
      p:= 0;
      for k:= 1 to 33 do
      begin
        m:= 0;
        for l:= 0 to high(crc[i]) do
          if length(crc[i][l].text)=k then inc(m);
        image2.canvas.brush.color:= colors[k];
        image2.canvas.pen.color:= colors[k];
        image2.canvas.Rectangle(
            (image2.width*I) div nbcrc, image2.height-((image2.height*p) div j),
            (image2.width*(I+1)) div nbcrc, image2.height-(image2.height*(p+m))div j);
        inc(p, m);
      end;
    end;

    // summary
    for i:= 0 to nbcrc-1 do
      crcd[i]:= length(crc[i]);

    application.MessageBox(pchar(
              inttostr(nbext)+' : entries'#13#10+
              inttostr((f.position-debdata) div 2)+' : bytes of data table'#13#10+
              inttostr((nbext*10+nbcrc*8+25) div 2)+' : bytes of index'#13#10+
              inttohex(s3[0].value, 5)+' : min entry'#13#10+
              inttohex(s3[nbext-1].value, 5)+' : max entry'#13#10+
              floattostr(Mean(crcd))+' : mean number of entry in each crc table'#13#10+
              floattostr(stddev(crcd))+' : stdev number of entry in each crc table'
              )
              , 'sumary', MB_OK);

    // rplobject header
    f.InsertInt($02BEE, 0);
    f.InsertInt(f.size, 5);

    // final save
    f.SaveToHPFile(SaveDialog2.filename);
    f.free;
    s.free;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  i: integer;
begin
  for i:= 0 to nbcrc-1 do
    listBox1.items.add(inttohex(i, 2));
end;

procedure TForm1.ListBox1Click(Sender: TObject);
var
  j: integer;
begin
  memo2.lines.Clear;
  for j:=0 to high(crc[listBox1.itemIndex]) do
    memo2.lines.add(crc[listBox1.itemIndex][j].text);
end;

end.
